@@ -16,6 +16,13 @@
 
 
 #include "bson-iter.h"
+#include "bson-config.h"
+#ifdef BSON_EXPERIMENTAL_FEATURES
+#include "bson-decimal128.h"
+#endif
+#ifdef HAVE_STRINGS_H
+# include <strings.h>
+#endif
 
 
 #define ITER_TYPE(i) ((bson_type_t) *((i)->raw + (i)->type))
@@ -41,8 +48,8 @@ bool
 bson_iter_init (bson_iter_t  *iter, /* OUT */
                 const bson_t *bson) /* IN */
 {
-   bson_return_val_if_fail (iter, false);
-   bson_return_val_if_fail (bson, false);
+   BSON_ASSERT (iter);
+   BSON_ASSERT (bson);
 
    if (BSON_UNLIKELY (bson->len < 5)) {
       memset (iter, 0, sizeof *iter);
@@ -89,8 +96,8 @@ bson_iter_recurse (const bson_iter_t *iter,  /* IN */
    const uint8_t *data = NULL;
    uint32_t len = 0;
 
-   bson_return_val_if_fail (iter, false);
-   bson_return_val_if_fail (child, false);
+   BSON_ASSERT (iter);
+   BSON_ASSERT (child);
 
    if (ITER_TYPE (iter) == BSON_TYPE_DOCUMENT) {
       bson_iter_document (iter, &len, &data);
@@ -138,9 +145,9 @@ bson_iter_init_find (bson_iter_t  *iter, /* INOUT */
                      const bson_t *bson, /* IN */
                      const char   *key)  /* IN */
 {
-   bson_return_val_if_fail (iter, false);
-   bson_return_val_if_fail (bson, false);
-   bson_return_val_if_fail (key, false);
+   BSON_ASSERT (iter);
+   BSON_ASSERT (bson);
+   BSON_ASSERT (key);
 
    return bson_iter_init (iter, bson) && bson_iter_find (iter, key);
 }
@@ -167,9 +174,9 @@ bson_iter_init_find_case (bson_iter_t  *iter, /* INOUT */
                           const bson_t *bson, /* IN */
                           const char   *key)  /* IN */
 {
-   bson_return_val_if_fail (iter, false);
-   bson_return_val_if_fail (bson, false);
-   bson_return_val_if_fail (key, false);
+   BSON_ASSERT (iter);
+   BSON_ASSERT (bson);
+   BSON_ASSERT (key);
 
    return bson_iter_init (iter, bson) && bson_iter_find_case (iter, key);
 }
@@ -198,9 +205,6 @@ _bson_iter_find_with_len (bson_iter_t *iter,   /* INOUT */
 {
    const char *ikey;
 
-   bson_return_val_if_fail (iter, false);
-   bson_return_val_if_fail (key, false);
-
    if (keylen == 0) {
       return false;
    }
@@ -243,8 +247,8 @@ bool
 bson_iter_find (bson_iter_t *iter, /* INOUT */
                 const char  *key)  /* IN */
 {
-   bson_return_val_if_fail (iter, false);
-   bson_return_val_if_fail (key, false);
+   BSON_ASSERT (iter);
+   BSON_ASSERT (key);
 
    return _bson_iter_find_with_len (iter, key, -1);
 }
@@ -272,8 +276,8 @@ bool
 bson_iter_find_case (bson_iter_t *iter, /* INOUT */
                      const char  *key)  /* IN */
 {
-   bson_return_val_if_fail (iter, false);
-   bson_return_val_if_fail (key, false);
+   BSON_ASSERT (iter);
+   BSON_ASSERT (key);
 
    while (bson_iter_next (iter)) {
 #ifdef BSON_OS_WIN32
@@ -316,9 +320,9 @@ bson_iter_find_descendant (bson_iter_t *iter,       /* INOUT */
    const char *dot;
    size_t sublen;
 
-   bson_return_val_if_fail (iter, false);
-   bson_return_val_if_fail (dotkey, false);
-   bson_return_val_if_fail (descendant, false);
+   BSON_ASSERT (iter);
+   BSON_ASSERT (dotkey);
+   BSON_ASSERT (descendant);
 
    if ((dot = strchr (dotkey, '.'))) {
       sublen = dot - dotkey;
@@ -363,7 +367,7 @@ bson_iter_find_descendant (bson_iter_t *iter,       /* INOUT */
 const char *
 bson_iter_key (const bson_iter_t *iter) /* IN */
 {
-   bson_return_val_if_fail (iter, NULL);
+   BSON_ASSERT (iter);
 
    return bson_iter_key_unsafe (iter);
 }
@@ -389,9 +393,9 @@ bson_iter_key (const bson_iter_t *iter) /* IN */
 bson_type_t
 bson_iter_type (const bson_iter_t *iter) /* IN */
 {
-   bson_return_val_if_fail (iter, BSON_TYPE_EOD);
-   bson_return_val_if_fail (iter->raw, BSON_TYPE_EOD);
-   bson_return_val_if_fail (iter->len, BSON_TYPE_EOD);
+   BSON_ASSERT (iter);
+   BSON_ASSERT (iter->raw);
+   BSON_ASSERT (iter->len);
 
    return bson_iter_type_unsafe (iter);
 }
@@ -400,34 +404,43 @@ bson_iter_type (const bson_iter_t *iter) /* IN */
 /*
  *--------------------------------------------------------------------------
  *
- * bson_iter_next --
+ * _bson_iter_next_internal --
  *
- *       Advances @iter to the next field of the underlying BSON document.
- *       If all fields have been exhausted, then %false is returned.
+ *       Internal function to advance @iter to the next field and retrieve
+ *       the key and BSON type before error-checking.
  *
- *       It is a programming error to use @iter after this function has
- *       returned false.
- *
- * Returns:
- *       true if the iter was advanced to the next record.
- *       otherwise false and @iter should be considered invalid.
+ * Return:
+ *       true if an element was decoded, else false.
  *
  * Side effects:
- *       @iter may be invalidated.
+ *       @key and @bson_type are set.
+ *
+ *       If the return value is false:
+ *        - @iter is invalidated: @iter->raw is NULLed
+ *        - @unsupported is set to true if the bson type is unsupported
+ *        - otherwise if the BSON is corrupt, @iter->err_off is nonzero
+ *        - otherwise @bson_type is set to BSON_TYPE_EOD
  *
  *--------------------------------------------------------------------------
  */
 
-bool
-bson_iter_next (bson_iter_t *iter) /* INOUT */
+static bool
+_bson_iter_next_internal (bson_iter_t  *iter,         /* INOUT */
+                          const char  **key,          /* OUT */
+                          uint32_t     *bson_type,    /* OUT */
+                          bool         *unsupported)  /* OUT */
 {
    const uint8_t *data;
    uint32_t o;
    unsigned int len;
 
-   bson_return_val_if_fail (iter, false);
+   BSON_ASSERT (iter);
+
+   *unsupported = false;
 
    if (!iter->raw) {
+      *key = NULL;
+      *bson_type = BSON_TYPE_EOD;
       return false;
    }
 
@@ -453,7 +466,10 @@ bson_iter_next (bson_iter_t *iter) /* INOUT */
 
 fill_data_fields:
 
-   switch (ITER_TYPE (iter)) {
+   *key = bson_iter_key_unsafe (iter);
+   *bson_type = ITER_TYPE (iter);
+
+   switch (*bson_type) {
    case BSON_TYPE_DATE_TIME:
    case BSON_TYPE_DOUBLE:
    case BSON_TYPE_INT64:
@@ -668,6 +684,11 @@ fill_data_fields:
    case BSON_TYPE_INT32:
       iter->next_off = o + 4;
       break;
+#ifdef BSON_EXPERIMENTAL_FEATURES
+   case BSON_TYPE_DECIMAL128:
+      iter->next_off = o + 16;
+      break;
+#endif
    case BSON_TYPE_MAXKEY:
    case BSON_TYPE_MINKEY:
    case BSON_TYPE_NULL:
@@ -675,8 +696,10 @@ fill_data_fields:
       iter->d1 = -1;
       iter->next_off = o;
       break;
-   case BSON_TYPE_EOD:
    default:
+      *unsupported = true;
+      /* FALL THROUGH */
+   case BSON_TYPE_EOD:
       iter->err_off = o;
       goto mark_invalid;
    }
@@ -704,6 +727,38 @@ mark_invalid:
 }
 
 
+/*
+ *--------------------------------------------------------------------------
+ *
+ * bson_iter_next --
+ *
+ *       Advances @iter to the next field of the underlying BSON document.
+ *       If all fields have been exhausted, then %false is returned.
+ *
+ *       It is a programming error to use @iter after this function has
+ *       returned false.
+ *
+ * Returns:
+ *       true if the iter was advanced to the next record.
+ *       otherwise false and @iter should be considered invalid.
+ *
+ * Side effects:
+ *       @iter may be invalidated.
+ *
+ *--------------------------------------------------------------------------
+ */
+
+bool
+bson_iter_next (bson_iter_t *iter) /* INOUT */
+{
+   uint32_t bson_type;
+   const char *key;
+   bool unsupported;
+
+   return _bson_iter_next_internal (iter, &key, &bson_type, &unsupported);
+}
+
+
 /*
  *--------------------------------------------------------------------------
  *
@@ -713,7 +768,7 @@ mark_invalid:
  *       @subtype.  The length of @binary in bytes is stored in @binary_len.
  *
  *       @binary should not be modified or freed and is only valid while
- *       @iter is on the current field.
+ *       @iter's bson_t is valid and unmodified.
  *
  * Parameters:
  *       @iter: A bson_iter_t
@@ -738,8 +793,8 @@ bson_iter_binary (const bson_iter_t  *iter,        /* IN */
 {
    bson_subtype_t backup;
 
-   bson_return_if_fail (iter);
-   bson_return_if_fail (!binary || binary_len);
+   BSON_ASSERT (iter);
+   BSON_ASSERT (!binary || binary_len);
 
    if (ITER_TYPE (iter) == BSON_TYPE_BINARY) {
       if (!subtype) {
@@ -795,7 +850,7 @@ bson_iter_binary (const bson_iter_t  *iter,        /* IN */
 bool
 bson_iter_bool (const bson_iter_t *iter) /* IN */
 {
-   bson_return_val_if_fail (iter, 0);
+   BSON_ASSERT (iter);
 
    if (ITER_TYPE (iter) == BSON_TYPE_BOOL) {
       return bson_iter_bool_unsafe (iter);
@@ -828,7 +883,7 @@ bson_iter_bool (const bson_iter_t *iter) /* IN */
 bool
 bson_iter_as_bool (const bson_iter_t *iter) /* IN */
 {
-   bson_return_val_if_fail (iter, 0);
+   BSON_ASSERT (iter);
 
    switch ((int)ITER_TYPE (iter)) {
    case BSON_TYPE_BOOL:
@@ -869,7 +924,7 @@ bson_iter_as_bool (const bson_iter_t *iter) /* IN */
 double
 bson_iter_double (const bson_iter_t *iter) /* IN */
 {
-   bson_return_val_if_fail (iter, 0);
+   BSON_ASSERT (iter);
 
    if (ITER_TYPE (iter) == BSON_TYPE_DOUBLE) {
       return bson_iter_double_unsafe (iter);
@@ -898,7 +953,7 @@ bson_iter_double (const bson_iter_t *iter) /* IN */
 int32_t
 bson_iter_int32 (const bson_iter_t *iter) /* IN */
 {
-   bson_return_val_if_fail (iter, 0);
+   BSON_ASSERT (iter);
 
    if (ITER_TYPE (iter) == BSON_TYPE_INT32) {
       return bson_iter_int32_unsafe (iter);
@@ -928,7 +983,7 @@ bson_iter_int32 (const bson_iter_t *iter) /* IN */
 int64_t
 bson_iter_int64 (const bson_iter_t *iter) /* IN */
 {
-   bson_return_val_if_fail (iter, 0);
+   BSON_ASSERT (iter);
 
    if (ITER_TYPE (iter) == BSON_TYPE_INT64) {
       return bson_iter_int64_unsafe (iter);
@@ -962,7 +1017,7 @@ bson_iter_int64 (const bson_iter_t *iter) /* IN */
 int64_t
 bson_iter_as_int64 (const bson_iter_t *iter) /* IN */
 {
-   bson_return_val_if_fail (iter, 0);
+   BSON_ASSERT (iter);
 
    switch ((int)ITER_TYPE (iter)) {
    case BSON_TYPE_BOOL:
@@ -979,6 +1034,40 @@ bson_iter_as_int64 (const bson_iter_t *iter) /* IN */
 }
 
 
+#ifdef BSON_EXPERIMENTAL_FEATURES
+/*
+ *--------------------------------------------------------------------------
+ *
+ * bson_iter_decimal128 --
+ *
+ *       This function retrieves the current field of type %BSON_TYPE_DECIMAL128.
+ *       The result is valid while @iter is valid, and is stored in @dec.
+ *
+ * Returns:
+ *
+ *       True on success, false on failure.
+ *
+ * Side Effects:
+ *    None.
+ *
+ *--------------------------------------------------------------------------
+ */
+bool
+bson_iter_decimal128 (const bson_iter_t *iter,      /* IN */
+                      bson_decimal128_t *dec)       /* OUT */
+{
+   BSON_ASSERT (iter);
+
+   if (ITER_TYPE (iter) == BSON_TYPE_DECIMAL128) {
+      bson_iter_decimal128_unsafe (iter, dec);
+      return true;
+   }
+
+   return false;
+}
+#endif
+
+
 /*
  *--------------------------------------------------------------------------
  *
@@ -999,7 +1088,7 @@ bson_iter_as_int64 (const bson_iter_t *iter) /* IN */
 const bson_oid_t *
 bson_iter_oid (const bson_iter_t *iter) /* IN */
 {
-   bson_return_val_if_fail (iter, NULL);
+   BSON_ASSERT (iter);
 
    if (ITER_TYPE (iter) == BSON_TYPE_OID) {
       return bson_iter_oid_unsafe (iter);
@@ -1033,7 +1122,7 @@ bson_iter_regex (const bson_iter_t *iter,    /* IN */
    const char *ret = NULL;
    const char *ret_options = NULL;
 
-   bson_return_val_if_fail (iter, NULL);
+   BSON_ASSERT (iter);
 
    if (ITER_TYPE (iter) == BSON_TYPE_REGEX) {
       ret = (const char *)(iter->raw + iter->d1);
@@ -1073,7 +1162,7 @@ const char *
 bson_iter_utf8 (const bson_iter_t *iter,   /* IN */
                 uint32_t          *length) /* OUT */
 {
-   bson_return_val_if_fail (iter, NULL);
+   BSON_ASSERT (iter);
 
    if (ITER_TYPE (iter) == BSON_TYPE_UTF8) {
       if (length) {
@@ -1117,7 +1206,7 @@ bson_iter_dup_utf8 (const bson_iter_t *iter,   /* IN */
    const char *str;
    char *ret = NULL;
 
-   bson_return_val_if_fail (iter, NULL);
+   BSON_ASSERT (iter);
 
    if ((str = bson_iter_utf8 (iter, &local_length))) {
       ret = bson_malloc0 (local_length + 1);
@@ -1159,7 +1248,7 @@ const char *
 bson_iter_code (const bson_iter_t *iter,   /* IN */
                 uint32_t          *length) /* OUT */
 {
-   bson_return_val_if_fail (iter, NULL);
+   BSON_ASSERT (iter);
 
    if (ITER_TYPE (iter) == BSON_TYPE_CODE) {
       if (length) {
@@ -1212,7 +1301,7 @@ bson_iter_codewscope (const bson_iter_t  *iter,      /* IN */
 {
    uint32_t len;
 
-   bson_return_val_if_fail (iter, NULL);
+   BSON_ASSERT (iter);
 
    if (ITER_TYPE (iter) == BSON_TYPE_CODEWSCOPE) {
       if (length) {
@@ -1278,7 +1367,7 @@ bson_iter_dbpointer (const bson_iter_t  *iter,           /* IN */
                      const char        **collection,     /* OUT */
                      const bson_oid_t  **oid)            /* OUT */
 {
-   bson_return_if_fail (iter);
+   BSON_ASSERT (iter);
 
    if (collection) {
       *collection = NULL;
@@ -1338,7 +1427,7 @@ bson_iter_symbol (const bson_iter_t *iter,   /* IN */
    const char *ret = NULL;
    uint32_t ret_length = 0;
 
-   bson_return_val_if_fail (iter, NULL);
+   BSON_ASSERT (iter);
 
    if (ITER_TYPE (iter) == BSON_TYPE_SYMBOL) {
       ret = (const char *)(iter->raw + iter->d2);
@@ -1373,7 +1462,7 @@ bson_iter_symbol (const bson_iter_t *iter,   /* IN */
 int64_t
 bson_iter_date_time (const bson_iter_t *iter) /* IN */
 {
-   bson_return_val_if_fail (iter, 0);
+   BSON_ASSERT (iter);
 
    if (ITER_TYPE (iter) == BSON_TYPE_DATE_TIME) {
       return bson_iter_int64_unsafe (iter);
@@ -1403,7 +1492,7 @@ bson_iter_date_time (const bson_iter_t *iter) /* IN */
 time_t
 bson_iter_time_t (const bson_iter_t *iter) /* IN */
 {
-   bson_return_val_if_fail (iter, 0);
+   BSON_ASSERT (iter);
 
    if (ITER_TYPE (iter) == BSON_TYPE_DATE_TIME) {
       return bson_iter_time_t_unsafe (iter);
@@ -1444,7 +1533,7 @@ bson_iter_timestamp (const bson_iter_t *iter,      /* IN */
    uint32_t ret_timestamp = 0;
    uint32_t ret_increment = 0;
 
-   bson_return_if_fail (iter);
+   BSON_ASSERT (iter);
 
    if (ITER_TYPE (iter) == BSON_TYPE_TIMESTAMP) {
       memcpy (&encoded, iter->raw + iter->d1, sizeof (encoded));
@@ -1488,7 +1577,7 @@ void
 bson_iter_timeval (const bson_iter_t *iter,  /* IN */
                    struct timeval    *tv)    /* OUT */
 {
-   bson_return_if_fail (iter);
+   BSON_ASSERT (iter);
 
    if (ITER_TYPE (iter) == BSON_TYPE_DATE_TIME) {
       bson_iter_timeval_unsafe (iter, tv);
@@ -1547,9 +1636,9 @@ bson_iter_document (const bson_iter_t  *iter,         /* IN */
                     uint32_t           *document_len, /* OUT */
                     const uint8_t     **document)     /* OUT */
 {
-   bson_return_if_fail (iter);
-   bson_return_if_fail (document_len);
-   bson_return_if_fail (document);
+   BSON_ASSERT (iter);
+   BSON_ASSERT (document_len);
+   BSON_ASSERT (document);
 
    *document = NULL;
    *document_len = 0;
@@ -1609,9 +1698,9 @@ bson_iter_array (const bson_iter_t  *iter,      /* IN */
                  uint32_t           *array_len, /* OUT */
                  const uint8_t     **array)     /* OUT */
 {
-   bson_return_if_fail (iter);
-   bson_return_if_fail (array_len);
-   bson_return_if_fail (array);
+   BSON_ASSERT (iter);
+   BSON_ASSERT (array_len);
+   BSON_ASSERT (array);
 
    *array = NULL;
    *array_len = 0;
@@ -1646,6 +1735,7 @@ bson_iter_array (const bson_iter_t  *iter,      /* IN */
 #define VISIT_INT32 VISIT_FIELD (int32)
 #define VISIT_TIMESTAMP VISIT_FIELD (timestamp)
 #define VISIT_INT64 VISIT_FIELD (int64)
+#define VISIT_DECIMAL128 VISIT_FIELD (decimal128)
 #define VISIT_MAXKEY VISIT_FIELD (maxkey)
 #define VISIT_MINKEY VISIT_FIELD (minkey)
 
@@ -1689,24 +1779,24 @@ bson_iter_visit_all (bson_iter_t          *iter,    /* INOUT */
                      const bson_visitor_t *visitor, /* IN */
                      void                 *data)    /* IN */
 {
+   uint32_t bson_type;
    const char *key;
+   bool unsupported;
 
-   bson_return_val_if_fail (iter, false);
-   bson_return_val_if_fail (visitor, false);
-
-   while (bson_iter_next (iter)) {
-      key = bson_iter_key_unsafe (iter);
+   BSON_ASSERT (iter);
+   BSON_ASSERT (visitor);
 
+   while (_bson_iter_next_internal (iter, &key, &bson_type, &unsupported)) {
       if (*key && !bson_utf8_validate (key, strlen (key), false)) {
          iter->err_off = iter->off;
-         return true;
+         break;
       }
 
       if (VISIT_BEFORE (iter, key, data)) {
          return true;
       }
 
-      switch (bson_iter_type (iter)) {
+      switch (bson_type) {
       case BSON_TYPE_DOUBLE:
 
          if (VISIT_DOUBLE (iter, key, bson_iter_double (iter), data)) {
@@ -1897,6 +1987,18 @@ bson_iter_visit_all (bson_iter_t          *iter,    /* INOUT */
          }
 
          break;
+#ifdef BSON_EXPERIMENTAL_FEATURES
+      case BSON_TYPE_DECIMAL128:
+         {
+            bson_decimal128_t dec;
+            bson_iter_decimal128 (iter, &dec);
+
+            if (VISIT_DECIMAL128 (iter, key, &dec, data)) {
+               return true;
+            }
+         }
+         break;
+#endif
       case BSON_TYPE_MAXKEY:
 
          if (VISIT_MAXKEY (iter, bson_iter_key_unsafe (iter), data)) {
@@ -1922,6 +2024,13 @@ bson_iter_visit_all (bson_iter_t          *iter,    /* INOUT */
    }
 
    if (iter->err_off) {
+      if (unsupported &&
+            visitor->visit_unsupported_type &&
+            bson_utf8_validate (key, strlen (key), false)) {
+         visitor->visit_unsupported_type (iter, key, bson_type, data);
+         return false;
+      }
+
       VISIT_CORRUPT (iter, data);
    }
 
@@ -1952,8 +2061,8 @@ void
 bson_iter_overwrite_bool (bson_iter_t *iter,  /* IN */
                           bool         value) /* IN */
 {
-   bson_return_if_fail (iter);
-   bson_return_if_fail (value == 1 || value == 0);
+   BSON_ASSERT (iter);
+   value = !!value;
 
    if (ITER_TYPE (iter) == BSON_TYPE_BOOL) {
       memcpy ((void *)(iter->raw + iter->d1), &value, 1);
@@ -1982,7 +2091,7 @@ void
 bson_iter_overwrite_int32 (bson_iter_t *iter,  /* IN */
                            int32_t      value) /* IN */
 {
-   bson_return_if_fail (iter);
+   BSON_ASSERT (iter);
 
    if (ITER_TYPE (iter) == BSON_TYPE_INT32) {
 #if BSON_BYTE_ORDER != BSON_LITTLE_ENDIAN
@@ -2014,7 +2123,7 @@ void
 bson_iter_overwrite_int64 (bson_iter_t *iter,   /* IN */
                            int64_t      value)  /* IN */
 {
-   bson_return_if_fail (iter);
+   BSON_ASSERT (iter);
 
    if (ITER_TYPE (iter) == BSON_TYPE_INT64) {
 #if BSON_BYTE_ORDER != BSON_LITTLE_ENDIAN
@@ -2046,7 +2155,7 @@ void
 bson_iter_overwrite_double (bson_iter_t *iter,  /* IN */
                             double       value) /* IN */
 {
-   bson_return_if_fail (iter);
+   BSON_ASSERT (iter);
 
    if (ITER_TYPE (iter) == BSON_TYPE_DOUBLE) {
       value = BSON_DOUBLE_TO_LE (value);
@@ -2055,6 +2164,43 @@ bson_iter_overwrite_double (bson_iter_t *iter,  /* IN */
 }
 
 
+#ifdef BSON_EXPERIMENTAL_FEATURES
+/*
+ *--------------------------------------------------------------------------
+ *
+ * bson_iter_overwrite_decimal128 --
+ *
+ *       Overwrites the current BSON_TYPE_DECIMAL128 field with a new value.
+ *       This is performed in-place and therefore no keys are moved.
+ *
+ * Returns:
+ *    None.
+ *
+ * Side effects:
+ *    None.
+ *
+ *--------------------------------------------------------------------------
+ */
+void
+bson_iter_overwrite_decimal128 (bson_iter_t       *iter,   /* IN */
+                                bson_decimal128_t *value)  /* IN */
+{
+   BSON_ASSERT (iter);
+
+   if (ITER_TYPE (iter) == BSON_TYPE_DECIMAL128) {
+#if BSON_BYTE_ORDER != BSON_LITTLE_ENDIAN
+      uint64_t data[2];
+      data[0] = BSON_UINT64_TO_LE (value->low);
+      data[1] = BSON_UINT64_TO_LE (value->high);
+      memcpy ((void *)(iter->raw + iter->d1), data, sizeof (data));
+#else
+      memcpy ((void *)(iter->raw + iter->d1), value, sizeof (*value));
+#endif
+   }
+}
+#endif
+
+
 /*
  *--------------------------------------------------------------------------
  *
@@ -2079,7 +2225,7 @@ bson_iter_value (bson_iter_t *iter) /* IN */
 {
    bson_value_t *value;
 
-   bson_return_val_if_fail (iter, NULL);
+   BSON_ASSERT (iter);
 
    value = &iter->value;
    value->value_type = ITER_TYPE (iter);
@@ -2163,6 +2309,11 @@ bson_iter_value (bson_iter_t *iter) /* IN */
    case BSON_TYPE_INT64:
       value->value.v_int64 = bson_iter_int64 (iter);
       break;
+#ifdef BSON_EXPERIMENTAL_FEATURES
+   case BSON_TYPE_DECIMAL128:
+      bson_iter_decimal128 (iter, &(value->value.v_decimal128));
+      break;
+#endif
    case BSON_TYPE_NULL:
    case BSON_TYPE_UNDEFINED:
    case BSON_TYPE_MAXKEY:
