@@ -78,12 +78,12 @@ void                          mongoc_bulk_operation_set_collection    (mongoc_bu
                                                                        const char                    *collection);
 void                          mongoc_bulk_operation_set_client        (mongoc_bulk_operation_t       *bulk,
                                                                        void                          *client);
+/* These names include the term "hint" for backward compatibility, should be
+ * mongoc_bulk_operation_get_server_id, mongoc_bulk_operation_set_server_id. */
 void                          mongoc_bulk_operation_set_hint          (mongoc_bulk_operation_t       *bulk,
-                                                                       uint32_t                       hint);
-void                          mongoc_bulk_operation_set_flags         (mongoc_bulk_operation_t       *bulk,
-                                                                       mongoc_bulk_write_flags_t      flags);
+                                                                       uint32_t                       server_id);
+uint32_t                      mongoc_bulk_operation_get_hint          (const mongoc_bulk_operation_t *bulk);
 const mongoc_write_concern_t *mongoc_bulk_operation_get_write_concern (const mongoc_bulk_operation_t *bulk);
-
 BSON_END_DECLS
 
 
