@@ -17,7 +17,7 @@
 #ifndef MONGOC_COUNTERS_PRIVATE_H
 #define MONGOC_COUNTERS_PRIVATE_H
 
-#if !defined (MONGOC_I_AM_A_DRIVER) && !defined (MONGOC_COMPILATION)
+#if !defined (MONGOC_COMPILATION)
 #error "Only <mongoc.h> can be included directly."
 #endif
 
@@ -40,6 +40,7 @@ BSON_BEGIN_DECLS
 
 
 void _mongoc_counters_init (void);
+void _mongoc_counters_cleanup (void);
 
 
 static BSON_INLINE unsigned
