@@ -59,6 +59,7 @@ mongoc_database_t *
 _mongoc_database_new (mongoc_client_t              *client,
                       const char                   *name,
                       const mongoc_read_prefs_t    *read_prefs,
+                      const mongoc_read_concern_t  *read_concern,
                       const mongoc_write_concern_t *write_concern)
 {
    mongoc_database_t *db;
@@ -73,6 +74,9 @@ _mongoc_database_new (mongoc_client_t              *client,
    db->write_concern = write_concern ?
       mongoc_write_concern_copy(write_concern) :
       mongoc_write_concern_new();
+   db->read_concern = read_concern ?
+      mongoc_read_concern_copy(read_concern) :
+      mongoc_read_concern_new();
    db->read_prefs = read_prefs ?
       mongoc_read_prefs_copy(read_prefs) :
       mongoc_read_prefs_new(MONGOC_READ_PRIMARY);
@@ -111,6 +115,11 @@ mongoc_database_destroy (mongoc_database_t *database)
       database->read_prefs = NULL;
    }
 
+   if (database->read_concern) {
+      mongoc_read_concern_destroy(database->read_concern);
+      database->read_concern = NULL;
+   }
+
    if (database->write_concern) {
       mongoc_write_concern_destroy(database->write_concern);
       database->write_concern = NULL;
@@ -121,13 +130,41 @@ mongoc_database_destroy (mongoc_database_t *database)
    EXIT;
 }
 
+/*
+ *--------------------------------------------------------------------------
+ *
+ * mongoc_database_copy --
+ *
+ *       Returns a copy of @database that needs to be freed by calling
+ *       mongoc_database_destroy.
+ *
+ * Returns:
+ *       A copy of this database.
+ *
+ * Side effects:
+ *       None.
+ *
+ *--------------------------------------------------------------------------
+ */
+
+mongoc_database_t *
+mongoc_database_copy (mongoc_database_t *database)
+{
+   ENTRY;
+
+   BSON_ASSERT (database);
+
+   RETURN(_mongoc_database_new(database->client, database->name,
+                               database->read_prefs, database->read_concern,
+                               database->write_concern));
+}
 
 mongoc_cursor_t *
 mongoc_database_command (mongoc_database_t         *database,
                          mongoc_query_flags_t       flags,
-                         uint32_t              skip,
-                         uint32_t              limit,
-                         uint32_t              batch_size,
+                         uint32_t                   skip,
+                         uint32_t                   limit,
+                         uint32_t                   batch_size,
                          const bson_t              *command,
                          const bson_t              *fields,
                          const mongoc_read_prefs_t *read_prefs)
@@ -135,10 +172,12 @@ mongoc_database_command (mongoc_database_t         *database,
    BSON_ASSERT (database);
    BSON_ASSERT (command);
 
-   if (!read_prefs) {
-      read_prefs = database->read_prefs;
-   }
-
+   /* Server Selection Spec: "The generic command method has a default read
+    * preference of mode 'primary'. The generic command method MUST ignore any
+    * default read preference from client, database or collection
+    * configuration. The generic command method SHOULD allow an optional read
+    * preference argument."
+    */
    return mongoc_client_command (database->client, database->name, flags, skip,
                                  limit, batch_size, command, fields, read_prefs);
 }
@@ -154,10 +193,12 @@ mongoc_database_command_simple (mongoc_database_t         *database,
    BSON_ASSERT (database);
    BSON_ASSERT (command);
 
-   if (!read_prefs) {
-      read_prefs = database->read_prefs;
-   }
-
+   /* Server Selection Spec: "The generic command method has a default read
+    * preference of mode 'primary'. The generic command method MUST ignore any
+    * default read preference from client, database or collection
+    * configuration. The generic command method SHOULD allow an optional read
+    * preference argument."
+    */
    return mongoc_client_command_simple (database->client, database->name,
                                         command, read_prefs, reply, error);
 }
@@ -514,6 +555,64 @@ mongoc_database_set_read_prefs (mongoc_database_t         *database,
 }
 
 
+/*
+ *--------------------------------------------------------------------------
+ *
+ * mongoc_database_get_read_concern --
+ *
+ *       Fetches the read concern for @database.
+ *
+ * Returns:
+ *       A mongoc_read_concern_t that should not be modified or freed.
+ *
+ * Side effects:
+ *       None.
+ *
+ *--------------------------------------------------------------------------
+ */
+
+const mongoc_read_concern_t *
+mongoc_database_get_read_concern (const mongoc_database_t *database)
+{
+   BSON_ASSERT (database);
+
+   return database->read_concern;
+}
+
+
+/*
+ *--------------------------------------------------------------------------
+ *
+ * mongoc_database_set_read_concern --
+ *
+ *       Set the default read concern for @database.
+ *
+ * Returns:
+ *       None.
+ *
+ * Side effects:
+ *       None.
+ *
+ *--------------------------------------------------------------------------
+ */
+
+void
+mongoc_database_set_read_concern (mongoc_database_t            *database,
+                                  const mongoc_read_concern_t  *read_concern)
+{
+   BSON_ASSERT (database);
+
+   if (database->read_concern) {
+      mongoc_read_concern_destroy (database->read_concern);
+      database->read_concern = NULL;
+   }
+
+   if (read_concern) {
+      database->read_concern = mongoc_read_concern_copy (read_concern);
+   }
+}
+
+
 /*
  *--------------------------------------------------------------------------
  *
@@ -613,13 +712,16 @@ mongoc_database_has_collection (mongoc_database_t *database,
    BSON_APPEND_UTF8 (&filter, "name", name);
 
    cursor = mongoc_database_find_collections (database, &filter, error);
-
-   if (!cursor ||
-       (error &&
-        ((error->domain != 0) ||
-         (error->code != 0)))) {
+   
+   if (!cursor) {
       return ret;
    }
+   
+   if (error &&
+        ((error->domain != 0) ||
+         (error->code != 0))) {
+      GOTO (cleanup);
+   }
 
    while (mongoc_cursor_next (cursor, &doc)) {
       if (bson_iter_init (&col_iter, doc) &&
@@ -778,12 +880,11 @@ mongoc_database_find_collections (mongoc_database_t *database,
 
    read_prefs = mongoc_read_prefs_new (MONGOC_READ_PRIMARY);
 
-   cursor = mongoc_database_command (database, MONGOC_QUERY_SLAVE_OK, 0, 0, 0,
-                                     &cmd, NULL, read_prefs);
+   cursor = _mongoc_cursor_new (database->client, database->name,
+                                MONGOC_QUERY_SLAVE_OK, 0, 0, 0, true,
+                                NULL, NULL, NULL, NULL);
 
-   _mongoc_cursor_cursorid_init(cursor);
-
-   cursor->limit = 0;
+   _mongoc_cursor_cursorid_init (cursor, &cmd);
 
    if (_mongoc_cursor_cursorid_prime (cursor)) {
        /* intentionally empty */
@@ -800,15 +901,6 @@ mongoc_database_find_collections (mongoc_database_t *database,
          } else if (error) {
             memcpy (error, &lerror, sizeof *error);
          }
-      } else {
-         /* TODO: remove this branch for general release.  Only relevant for RC */
-         mongoc_cursor_destroy (cursor);
-         cursor = mongoc_database_command (database, MONGOC_QUERY_SLAVE_OK, 0, 0, 0,
-                                           &cmd, NULL, read_prefs);
-
-         _mongoc_cursor_array_init(cursor, "collections");
-
-         cursor->limit = 0;
       }
    }
 
@@ -943,7 +1035,7 @@ mongoc_database_create_collection (mongoc_database_t *database,
             bson_set_error (error,
                             MONGOC_ERROR_COMMAND,
                             MONGOC_ERROR_COMMAND_INVALID_ARG,
-                            "The \"size\" parameter requires {\"capped\": true}");
+                            "The \"max\" parameter requires {\"capped\": true}");
             return NULL;
          }
       }
@@ -1018,6 +1110,7 @@ mongoc_database_create_collection (mongoc_database_t *database,
                                            database->name,
                                            name,
                                            database->read_prefs,
+                                           database->read_concern,
                                            database->write_concern);
    }
 
@@ -1034,8 +1127,9 @@ mongoc_database_get_collection (mongoc_database_t *database,
    BSON_ASSERT (database);
    BSON_ASSERT (collection);
 
-   return mongoc_client_get_collection (database->client, database->name,
-                                        collection);
+   return _mongoc_collection_new (database->client, database->name, collection,
+                                  database->read_prefs, database->read_concern,
+                                  database->write_concern);
 }
 
 
